;ADLIB.ASM (C)1993 Johannes Bjerregaard. All rights reserved.
;------------------------------------------------------------
;This module provides the low-level interface to the hardware of the AdLib
;card needed to play MIDI data using the MIDIEMU driver. This driver is meant
;to be assembled and saved as a binary image, interfacing to it through the
;function at address 0.

include smfpcmds.inc
include version.inc

;----------------------------------------------------------------------------
;Macro      :  adl_outm
;Purpose    :  Send a value to the AdLib card
;Parameters :  AL = register       AH = value
;Returns    :  -
;Remarks    :  This is the speed-optimized version of adl_out. The value
;              is not sent directly to the hardware but kept in a cache of
;              the complete register map. When all messages for each frame
;              have been completed, the function hw_updateregspace is called
;              to update on the changes that have occurred since last call.
;Registers
; destroyed :  -
;----------------------------------------------------------------------------
include adlib.inc

IFDEF STANDALONE
.model tiny
ELSE
.model small,c
ENDIF

global hw_entry_point      :far              ;global for debugging purposes
global hw_emuinit          :proc
global hw_loadtune         :proc
global hw_noteon           :proc
global hw_drumon           :proc
global hw_noteoff          :proc
global hw_pbchange         :proc
global hw_setport          :proc
global hw_setpbsens        :proc
global hw_setmodvalue      :proc
global hw_setpbvalue       :proc
global hw_isr              :proc
global vcop                :byte
global sfxnum              :byte
.code

;----------------------------------------------------------------------------
;Procedure  :  hw_entry_point
;Purpose    :  Call a function numbered by the value in register BX
;Parameters :  BX = function number (MHC_xxxxx). Other registers vary
;              depending on function.
;
;Returns    :  -
;Remarks    :  This is the only address which is guaranteed not to change.
;              It must be located at CS:0.
;Registers
; destroyed :  -
;----------------------------------------------------------------------------
hw_entry_point proc  far
            push  ax
            push  bx
            push  cx
            push  dx
            push  di
            push  ds
            push  es
            push  si

IFNDEF STANDALONE
            push  ax
            mov   ax,@data
            mov   ds,ax
            pop   ax
ELSE
            push  cs
            pop   ds
ENDIF
            mov   es,tunedataseg
            shl   bx,1                ;BX to word index
            mov   bx,[func_table+bx]  ;get address word of desired function
            call  bx

            pop   si
            pop   es
            pop   ds
            pop   di
            pop   dx
            pop   cx
            pop   bx
            pop   ax
            ret
hw_entry_point endp

include instrum.inc                   ;instrument structure
include smfptune.inc                  ;tune structure


.data
            db    'SMFP hardware interface for AdLib(R) V1.0'
            db    '(C)1993 Johannes Bjerregaard.'

insmultab    label word
             nn=0
             rept 128
             dw   nn*type instrument
             nn=nn+1
             endm

fqdata       label word               ;MIDI key to AdLib frequency value
	          db   0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
	          db	0,0,0,0,0,0,0,0,88,1,109,1,131,1,154,1
	          db	178,1,204,1,231,1,4,2,35,2,68,2,102,2,139,2
	          db	88,5,109,5,131,5,154,5,178,5,204,5,231,5,4,6
	          db	35,6,68,6,102,6,139,6,88,9,109,9,131,9,154,9
	          db	178,9,204,9,231,9,4,10,35,10,68,10,102,10,139,10
	          db	88,13,109,13,131,13,154,13,178,13,204,13,231,13,4,14
	          db	35,14,68,14,102,14,139,14,88,17,109,17,131,17,154,17
	          db	178,17,204,17,231,17,4,18,35,18,68,18,102,18,139,18
	          db	88,21,109,21,131,21,154,21,178,21,204,21,231,21,4,22
	          db	35,22,68,22,102,22,139,22,88,25,109,25,131,25,154,25
	          db	178,25,204,25,231,25,4,26,35,26,68,26,102,26,139,26
	          db	88,29,109,29,131,29,154,29,178,29,204,29,231,29,4,30
	          db	35,30,68,30,102,30,139,30
;            db   0,0,0,0,0,0,0,0
;	          db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
;	          db	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0


regspace     label byte   ;keeps a copy of the hardware registers
             db    256 dup (0)
;regwork      label byte   ;keeps a copy of what goes into the hardware regs
;             db    256 dup (0ffh)

func_table   label word
             dw    offset cs:hw_emuinit         ;initialize hardware.
             dw    offset cs:hw_loadtune
             dw    offset cs:hw_noteon          ;respond in hardware to MIDI note on.
             dw    offset cs:hw_drumon          ;same but only for channel 9.
             dw    offset cs:hw_noteoff         ;respond in hardware to MIDI note off.
             dw    offset cs:hw_pbchange        ;called pitch bend changes.
             dw    offset cs:hw_setport         ;called to set port to other than deflt.
             dw    offset cs:hw_setpbsens       ;called to change pitch bend sensitivity.
             dw    offset cs:hw_setpbvalue      ;called to change pitch bend position.
             dw    offset cs:hw_setmodvalue     ;called to change modulation position.


             dw    ?
;             dw    offset cs:hw_updateregspace  ;update hardware registers;
                                      ;called once per frame.


             dw    offset cs:hw_isr             ;interrupt subroutine - updates voice
                                      ;data.

AdLibIOAddr  label word
             dw         DEFAULT_IOPORT;port base address to use for AdLib I/O

velotab      label byte               ;translation table from MIDI velocity
             db    0,0,1,2,3,3,4,5    ;into AdLib volume data
             db    6,7,7,8,9,10,10,11
             db    12,13,14,14,15,16,17,17
             db    18,19,20,20,21,22,23,23
             db    24,25,25,26,27,28,28,29
             db    30,30,31,32,32,33,34,34
             db    35,36,36,37,38,38,39,39
             db    40,41,41,42,42,43,44,44
             db    45,45,46,46,47,47,48,48
             db    49,49,50,50,51,51,52,52
             db    53,53,54,54,54,55,55,56
             db    56,56,57,57,57,58,58,58
             db    59,59,59,59,60,60,60,61
             db    61,61,61,61,62,62,62,62
             db    62,62,63,63,63,63,63,63
             db    63,63,63,63,63,63,63,63

tunedataseg  label word               ;segment of tune info data
             dw         ?

hwvkey       label byte
             db  9 dup (?)            ;key number for each hardware voice

vcop         label byte
             db  0,1,2,8,9,10         ;operator offset table
             db  16,17,18

hwreg40      label byte
             db  22 dup (0ffh)        ;contents of hardware register 40h

pbsens       label byte
             db  12                   ;pitch bend sensitivity

pbvalue      label word
             dw  9 dup (2000h)        ;pitch bend values (-2000h - 2000h)

modvalue     label byte
             db  9 dup (?)            ;modulator value (0 - 7fh)

hwvtime      label word
             dw  9 dup (?)            ;time since last note on

hwvmodlfocnt label byte
             db  9 dup (?)            ;modulation LFO counter

hwvfq        label word
             dw  9 dup (?)            ;base frequency of note including
                                      ;initial setting of pitch bend

hwvins       label byte
             db  9 dup (0ffh)         ;instrument used with voice.
                                      ;if (0ffh) it's invalid and will
                                      ;not be updated by the realtime function
global hwvins:byte
;
;Local temporary variables
;
modv         label word
             dw ?
keyindex     label word
             dw ?
detunev      label word
             dw ?


.code
;
;procedures assume DS = CS = @code = @data unless otherwise explied.
;

adl_outp proc
        push dx
        mov  dx,AdLibIOAddr
        out  dx,al

        push ax

        rept AdLibIN1Repeat
        in   al,dx
        endm

        mov  al,ah
        inc  dx
        out  dx,al
        dec  dx

        rept AdLibIN2Repeat
        in   al,dx
        endm

        pop  ax
        pop  dx
        ret
adl_outp endp



IFNDEF STANDALONE
testwrite   proc
            push  es
            push  cx
            push  di
            push  si
            push  ax
            cld

            mov   cx,80*4
            mov   ax,0b800h
            mov   es,ax
            lea   si,tunedataseg
            xor   di,di
            mov   ah,1eh
vbix:
            lodsb
            stosw
            loop  vbix

            pop   ax
            pop   si
            pop   di
            pop   cx
            pop   es
            ret
testwrite   endp

;----------------------------------------------------------------------------
;Procedure  :  hw_updateregspace
;Purpose    :  Updates the AdLib hardware registers with the values that
;              differ from the two register workspaces regspace and regwork.
;Parameters :  -
;Returns    :  -
;Remarks    :  The main reason for this is speed. The secondary reason is that
;              in a MIDI-file reading situation, notes may often have a very
;              short lifetime and much time would be wasted unless changes are
;              kept in a temporary space and updated regularly. The procedure
;              should be called at the end of each frame before returning
;              control to the main program.
;Registers
; destroyed :  AX,BX,CX,DX,SI,DI
;----------------------------------------------------------------------------
hw_updateregspace proc
            ret
            push   es
            push   ds
            pop    es
            mov    dx,AdLibIOAddr
            mov    bx,0f6h
;            lea    si,regwork+20h ; regwork masked out
;if this function is masked in again, regwork must be masked back on
            lea    di,regspace+20h
            mov    cx,0f6h-20h
            cld
@nxl:
            repz   cmpsb
            jz     @fin

            mov    ax,bx
            sub    ax,cx
            out    dx,al

            rept   AdLibIN1Repeat
            in     al,dx
            endm

		      inc	 dx
            mov    al,[si-1]
            mov    [di-1],al
		      out	 dx,al
            dec    dx

            rept   AdLibIN2Repeat
            in     al,dx
            endm

            and    cx,cx
            jnz    @nxl
@fin:
;            call   testwrite
            pop    es
            ret
hw_updateregspace endp


ENDIF


;----------------------------------------------------------------------------
;Procedure  :  adl_noteon
;Purpose    :  Turn on a voice using a specified instrument, key and velocity
;Parameters :  AL = instrument     AH = voice
;              CL = velocity       CH = key
;Returns    :  -
;Remarks    :  -
;Registers
; destroyed :  AX,BX,CX,DX,SI,DI,BP,ES
;----------------------------------------------------------------------------
adl_noteon  proc
            push  bp
            mov   bl,ah                        ;
            xor   bh,bh
            mov   si,bx                        ;SI = voice byte index
            xor   bx,bx                        ;clear BX for use as index
            mov   [si+hwvmodlfocnt],bl         ;reset modulation lfo counter
            mov   [si+hwvins],al               ;remember instrument for voice

            mov   bl,cl                        ;
            mov   cl,[bx+velotab]              ;translate velocity

            mov   bl,al                        ;element number
            shl   bl,1
            xor   bh,bh
            mov   di,[insmultab+bx]

            add   ch,es:[di+insdata+coarse_detune]
            mov   [si+hwvkey],ch               ;remember key for voice

            ;calculate address of data for selected instrument

            ;look up operator numbers

            mov   dh,[si+vcop]                 ;DH = operator 2 index
            mov   dl,dh                        ;
            add   dl,3                         ;DL = operator 1 index

            ;
            ;clear voice in sound hardware - it's done directly here
            ;

            mov   bx,si
            shl   bx,1
            cmp   [bx+hwvtime],0               ;just reset?
            jnz   @an_doreset
            xor   bx,bx
            jmp   @an_dontreset
@an_doreset:
            xor   bx,bx
            mov   ax,si
            add   al,0b0h                      ;reset KON
            adl_out

            mov   ax,0ff80h                    ;reset sustain + release
            add   al,dh
            adl_out                            ;operator 2
            add   al,3                         ;
            adl_out                            ;operator 1
            sub   al,20h                       ;reset attack + decay
            adl_out                            ;operator 1
            sub   al,3
            adl_out                            ;operator 2

            shl   si,1
            mov   [si+hwvtime],bx              ;reset time for note on
            shr   si,1
@an_dontreset:
            ;-------------------------------------
            ;set up hardware registers for carrier
            ;-------------------------------------

            mov   bp,cx                        ;save a copy of CX for later


            ;the value 1 is the index number of the carrier into the 2-
            ;element-sized array of each element in the instrument structure

            ;output level is based on the output level specified in the
            ;instrument and the velocity specified as a parameter, using
            ;the following formula:

            ;   o = b - ((vs * v) / 64);
            ;
            ;where b is the base output level of the operator, vs is the
            ;velocity sensitivity and v is the velocity parameter

            mov   al,es:[di+insdata+1+velo_sens]       ;get velocity sensitivity
            mul   cl                           ;multiply with velocity param
            shr   ax,6                         ;divide by 64
            mov   ah,es:[di+insdata+1+output_level]    ;base output level of instr
            sub   ah,al                        ;subtract velocity part

            mov   al,40h                       ;output level base register

            ;if resulting output level is >=64 (unsigned byte), set to zero.
            ;Note that values less than zero are also caught here.

            test  ah,al                        ;(note use of register data)
            jz    @car_nooutg64
            xor   ah,ah                        ;reset data value to maximum

@car_nooutg64:

            or    ah,es:[di+1+insdata+reg40]
            mov   bl,dl
            xor   bh,bh
            mov   [hwreg40+bx],ah

            ;---------------------------------------
            ;set up hardware registers for modulator
            ;---------------------------------------

            ;this is a copy of the above code except for the operator selec-
            ;tion and the index of zero into instrument data instead of 1.

            ;in the modulator version of the output level calculation, the
            ;algorithm used is only similar to the one used with the carrier
            ;when the operator combination method is FM. If the voice is
            ;running in parallel mode, though, this formula is used:
            ;
            ;   o = 3fh - (((3fh - b) * v) / 64);
            ;
            ;where b is the base output level of the operator and v is the
            ;velocity

            mov   cx,bp                        ;restore calculated linear
                                               ;velocity value in CL and
                                               ;key number in CH
            test  es:[di+insdata+regc0],1      ;using parallel or FM?
            jnz   @mod_par                     ;jump if parallel

            mov   al,es:[di+insdata+0+velo_sens];get velocity sensitivity
            mul   cl                           ;multiply with velocity param
            shr   ax,6                         ;divide by 64
            mov   ah,es:[di+insdata+0+output_level]      ;base output level of instr
            sub   ah,al                        ;subtract velocity part
            jmp   @mod_skipmove
@mod_par:                                      ;parallel output volume
            mov   al,3fh                       ;refer to formula above
            sub   al,es:[di+insdata+0+output_level]
            mul   cl
            shr   ax,6
            mov   ah,3fh
            sub   ah,al

@mod_skipmove:

            mov   al,40h                       ;output level base register

            ;if resulting output level is >=64, set to zero

            test  ah,al                        ;(note use of register data)
            jz    @mod_nooutg64
            xor   ah,ah                        ;reset data value to maximum

@mod_nooutg64:
            or    ah,es:[di+insdata+0+reg40]   ;key scale level = bit 6+7

            mov   bl,dh
            mov   [hwreg40+bx],ah

            mov  ax,si                         ;voice # in AX
            pop  bp

            ;NOTE: Execution continues in hw_pbchange

adl_noteon  endp

;-----------------------------------------------------------------
;---*  NOTE: DO NOT ADD CODE HERE - CS:IP passes through here *---
;-----------------------------------------------------------------

;----------------------------------------------------------------------------
;Procedure  :  hw_pbchange
;Purpose    :  Change the base frequency according to a new pitch bend value
;Parameters :  AL = hardware voice number to change pitch bend value for
;Returns    :  AX = new base frequency of voice
;Remarks    :  Uses the value of _pbvalue[AL] as the new value to calculate
;              pitch base frequency from.
;Registers
; destroyed :  AX,BX,CX,DX,SI,DI
;----------------------------------------------------------------------------
hw_pbchange proc
            xor  ah,ah
            mov  di,ax                         ;DI as voice byte offset
            shl  al,1
            mov  si,ax                         ;SI as voice word offset

            mov  ax,[si+pbvalue]               ;AX = new pitch bend value
            cmp  ax,0
            jnz  @pb_notzero

            mov  bl,[di+hwvkey]                ;get key number of voice
            xor  bh,bh
            shl  bl,1
            mov  ax,[fqdata+bx]
            jmp  @fq_inAX

@pb_notzero:
            jl   @pb_less0

            ;Calculate new value of frequency from positive pitch bend
            ;formula is:
            ;
            ;     n  = (sens * bend) / 2000h
            ;     m  = (sens * bend) % 2000h
            ;     w  = key + n
            ;     mf = fqdata[w+1]-fqdata[w]
            ;     if (mf>=680) mf -= 680
            ;     fq = fqtable[w] + (mf * m) / 2000h
            ;
            ;where:  n  is the number of semitones displaced by pitch bend
            ;        m  is the remaining pitch bend between the displaced
            ;           semitone and the next semitone
            ;        w  is the actual key to which the pitch bend extends
            ;        mf the modulus m, represented in terms of frequency data
            ;        fq is the resulting frequency data

            xor  bh,bh
            mov  bl,pbsens
            mul  bx

            push ax

            mov  al,ah
            mov  ah,dl
            shr  ax,5                          ;AX = AX / 2000h

            pop  dx

            and  dx,1fffh                      ;DX = AX % 2000h

            ;now AX contains n and DX contains m

            xor  bh,bh
            mov  bl,[di+hwvkey]                ;get key number of voice
            add  bl,al                         ;w now in BL

            shl  bl,1                          ;w must be word offset
            mov  cx,[fqdata+bx+2]
            sub  cx,[fqdata+bx]                ;mf now in CX
            cmp  cx,680
            jb   @mf_below680
            sub  cx,680

@mf_below680:
            mov  ax,dx                         ;m now in AX
            mul  cx                            ;multiply m with corrected mf

            mov  al,ah
            mov  ah,dl
            shr  ax,5                          ;AX div 2000h

            add  ax,[fqdata+bx]                ;frequency now in AX
            jmp  @fq_inAX
@pb_less0:
            ;while this applies to a negative pitch bend value, the code is
            ;almost the same.

            add  ax,2000h
;            neg  ax                            ;abs (_pbvalue[hwv]) in AX

            xor  bh,bh
            mov  bl,pbsens
            mul  bx

            push ax

            mov  al,ah                         ;AX div 2000h
            mov  ah,dl
            shr  ax,5

            pop  dx
            and  dx,1fffh                      ;remainder

            ;now AX contains n and DX contains m

            xor  bh,bh
            mov  bl,[di+hwvkey]                ;get MIDI key number of voice
            add  bl,al                         ;w now in BL

            shl  bl,1                          ;w must be word offset

            mov  cx,[fqdata+bx+2]
            sub  cx,[fqdata+bx]
            cmp  cx,680
            jb   @mf_below680b
            sub  cx,680

@mf_below680b:
            mov  ax,dx
            mul  cx

            mov  al,ah                         ;AX div 2000h
            mov  ah,dl
            shr  ax,5

            add  ax,[fqdata+bx-24]

@fq_inAX:
            mov  [hwvfq+si],ax                 ;store new frequency value
                                               ;for use by interrupt function
            ret
hw_pbchange endp



;----------------------------------------------------------------------------
;Procedure  :  hw_noteon
;Purpose    :  Passes the parameter values on to the adl_noteon procedure.
;Parameters :  AL = instrument     AH = voice
;              CL = velocity       CH = key
;Returns    :  -
;Remarks    :  See adl_noteon. Procedure is unnessecary but provides a
;              homogenous interface to the adl_noteon function together with
;              the function hw_drumon (see below).
;Registers
; destroyed :  AX,BX,CX,DX,SI,DI,BP,ES
;----------------------------------------------------------------------------
hw_noteon   proc
            jmp    adl_noteon
hw_noteon   endp



;----------------------------------------------------------------------------
;Procedure  :  hw_drumon
;Purpose    :  Translates the passed parameter values through a table that
;              converts a virtual key to a physical key and an instrument,
;              and passes these values on to the adl_noteon function.
;Parameters :                      AH = voice
;              CL = velocity       CH = key
;Returns    :  -
;Remarks    :  See adl_noteon.
;Registers
; destroyed :  AX,BX,CX,DX,SI,DI,BP,ES
;----------------------------------------------------------------------------
hw_drumon   proc
            mov    dl,ch
            xor    dh,dh
            mov    si,dx
            mov    ch,es:[drumnote+si]
            mov    al,es:[drumins+si]
            jmp    adl_noteon
hw_drumon   endp




;----------------------------------------------------------------------------
;Procedure  :  hw_noteoff
;Purpose    :  processes a note off message by clearing the KON bit of the
;              AdLib card and changing the sustain level and decay rate to
;              the parameters in the current instrument.
;Parameters :  AL = voice
;
;Returns    :  -
;Remarks    :  Loads ES with tunedataseg, segment of SMFP info block
;Registers
; destroyed :  AX,BX,CX,DX,SI,DI,BP,ES
;----------------------------------------------------------------------------
hw_noteoff  proc

            ;calculate address of data for selected instrument

            mov   bl,al                        ;element number
            xor   bh,bh
            mov   si,bx                        ;SI = voice byte index
            mov   bl,[hwvins+bx]               ;get instrument for voice

            mov   ax,type instrument           ;size of struc instrument
            mul   bl                           ;times instrument number
            mov   di,ax                        ;ES:DI now points to instrument

            ;look up operator numbers

            mov   dh,[si+vcop]                 ;DH = operator 2 index
            mov   dl,dh                        ;
            add   dl,3                         ;DL = operator 1 index

IFDEF KEYOFF_ADSR
            ;change decay and sustain to keyoff values and make sure the
            ;sustaining sound bit is set to off

            mov   al,60h
            add   al,dl                        ;operator 1, A/D
            mov   ah,es:[di+insdata+keyoff_reg60+1]
            adl_out

            add   al,20h                       ;operator 1, S/R
            mov   ah,es:[di+insdata+keyoff_reg80+1]
            adl_out

            sub   al,60h                       ;operator 1, Sustaining sound

            mov   ah,es:[di+insdata+reg20+1]
            and   ah,0ffh-20h                  ;clear sustaining sound bit
            adl_out

ELSE        ;KEYOFF_ADSR not defined
            mov   al,20h
            add   al,dl
            mov   ah,es:[di+insdata+reg20+1]
            and   ah,0ffh-20h
            adl_out
ENDIF
            ;do the same procedure with the modulator


IFDEF KEYOFF_ADSR
            mov   al,60h
            add   al,dh                        ;operator 2 A/D
            mov   ah,es:[di+insdata+keyoff_reg60+0]
            adl_out

            add   al,20h                       ;operator 2, S/R
            mov   ah,es:[di+insdata+keyoff_reg80+0]
            adl_out

            sub   al,60h                       ;operator 2, Sustaining sound
            mov   ah,es:[di+insdata+reg20+0]
            and   ah,0ffh-20h                  ;clear sustaining sound bit
            adl_out
ELSE
            mov   al,20h
            add   al,dh
            mov   ah,es:[di+insdata+reg20+0]
            and   ah,0ffh-20h
            adl_out
ENDIF
            ret
hw_noteoff  endp



;----------------------------------------------------------------------------
;Procedure  :  hw_initadl
;Purpose    :  Initialize AdLib hardware.
;Parameters :  -
;Returns    :  -
;Remarks    :  Changes the AdLib card into melodic mode and resets all
;              relevant sound registers. This function is only meant to
;              be called once, namely before any tune is played. This
;              function is called internally from hw_emuinit
;Registers
; destroyed :  -
;----------------------------------------------------------------------------
hw_initadl  proc
            mov   ax,1     ;clear test register
            call  adl_outp
            mov   ax,2001h ;wave select enable
            call  adl_outp

            mov   ax,0bdh  ;deep vib/am disable
            call  adl_outp
            mov   cx,9
            xor   bx,bx
            xor   ah,ah
@l9v:
            mov   al,cl
            add   al,0afh
            call  adl_outp
            loop  @l9v

            mov   cx,16h
@l18v:
            mov   ax,0ff1fh
            add   al,cl
            adl_out
            add   al,20h
            adl_out
            add   al,20h
            adl_out
            add   al,20h
            adl_out
            xor   ah,ah
            add   al,60h
            adl_out
            loop  @l18v
            mov   cx,100h/2
            xor   ax,ax
            dec   ax
            xor   bx,bx
@l100b:
            mov   word ptr [bx+regspace],ax
            inc   bx
            inc   bx
            loop  @l100b

            ret
hw_initadl  endp



;----------------------------------------------------------------------------
;Procedure  :  hw_emuinit
;Purpose    :  Initialize hardware and AdLib driver.
;Parameters :  -
;Returns    :  -
;Remarks    :  Changes the AdLib card into melodic mode and resets all
;              relevant sound registers.
;Registers
; destroyed :  -
;----------------------------------------------------------------------------
hw_emuinit  proc
            call  hw_initadl
            mov   dx,2
            mov   cx,9
            xor   si,si
            xor   bx,bx
@@he_vl:
            mov   [hwvtime+si],dx
            mov   [hwvins+bx],0ffh
            inc   si
            inc   si
            inc   bx
            loop  @@he_vl
            ret
hw_emuinit  endp



;----------------------------------------------------------------------------
;Procedure  :  hw_loadtune
;Purpose    :  Initialize data tables for a tune.
;Parameters :  AX = segment of tune info structure.
;Returns    :  CF = 1 on error, clear on success.
;Remarks    :  -
;Registers
; destroyed :  ES (set to value of AX)
;----------------------------------------------------------------------------
hw_loadtune proc

            ;
            ;ensure file compatibility
            ;

            mov   es,ax
            cmp   es:[id],'MS'         ;ID, 1st word
            jz    @ok1
@err:
            stc
            ret
@ok1:
            cmp   es:[id+2],'PF'         ;ID, 2nd word
            jnz   @err

            ;
            ;ensure version compatibility
            ;

            cmp   es:[tuneversion],DATAVERSION  ;only DATAVERSION is supported
            jnz   @err

            ;
            ;version ok. Initialize tunedataseg and offset tables.
            ;

            mov   tunedataseg,ax
            push  ax
            mov   al,es:[pitchbenddefault]
            call  hw_setpbsens
            pop   ax
            ret
hw_loadtune endp



;----------------------------------------------------------------------------
;Procedure  :  hw_setport
;Purpose    :  Set the base port of AdLib card.
;Parameters :  AX = port.
;Returns    :  -
;Remarks    :  The port number defaults to DEFAULT_IOPORT.
;Registers
; destroyed :  -
;----------------------------------------------------------------------------
hw_setport  proc
            mov   AdLibIOAddr,ax
            ret
hw_setport  endp




;----------------------------------------------------------------------------
;Procedure  :  hw_setpbsens
;Purpose    :  Change the sensitivity of the pitch bender
;Parameters :  AL = new sensitivity value, in semitones.
;Returns    :  -
;Remarks    :  This function is called externally when a change in pitch bend
;              is received through a MIDI message and internally when a tune
;              is initialized.
;Registers
; destroyed :  -
;----------------------------------------------------------------------------
hw_setpbsens proc
            mov   pbsens,al
            ret
hw_setpbsens endp




;----------------------------------------------------------------------------
;Procedure  :  hw_setpbvalue
;Purpose    :  Change the value of the pitch bender for a specified voice.
;Parameters :  CX = new value      AH = voice
;Returns    :  -
;Remarks    :  This function is called externally when a change in pitch bend
;              for any voice is, or may be, due.
;Registers
; destroyed :  BX
;----------------------------------------------------------------------------
hw_setpbvalue proc
            mov   bl,ah
            shl   bl,1
            xor   bh,bh
            mov   [bx+pbvalue],cx
            ret
hw_setpbvalue endp




;----------------------------------------------------------------------------
;Procedure  :  hw_setmodvalue
;Purpose    :  Change the value of the mod wheel for a specified voice
;Parameters :  AL = new value      AH = voice
;Returns    :  -
;Remarks    :  This function is called externally when a change in modulation
;              is received through a MIDI message and internally when a tune
;              is initialized.
;Registers
; destroyed :  BX
;----------------------------------------------------------------------------
hw_setmodvalue proc
            mov   bl,ah
            xor   bh,bh
            mov   [bx+modvalue],al
            ret
hw_setmodvalue endp





;----------------------------------------------------------------------------
;Procedure  :  hw_isr
;Purpose    :  Subroutine called to update run-time voice changes.
;Parameters :  -
;Returns    :  -
;Remarks    :  -
;Registers
; destroyed :  AX,BX,CX,DX,SI,DI
;----------------------------------------------------------------------------
hw_isr      proc
            xor   si,si                               ;voice byte index
IFDEF ENABLESFX
@voice_loop:
            cmp   [sfxnum+si],0ffh                    ;voice used by sfx?
            jz    @@vc_notsfx
ELSE
            jmp   @voice_loop
ENDIF
@@jmpskip:
            jmp   @@skipvoice
IFNDEF ENABLESFX
@voice_loop:
ENDIF
@@vc_notsfx:
            mov   bl,[hwvins+si]                      ;instrument number
            cmp   bl,0ffh                             ;invalid instrument?
            jz    @@jmpskip
            mov   al,type instrument                  ;sizeof (instrument)
            mul   bl
            mov   bx,ax                               ;ES:[BX+insdata] =
                                                      ;data for instrument

            ;if hwvtime == 1, set up instrument data on hardware. This is
            ;not done until 1 frame after the note on occurs, because the
            ;same voice may be replaced several times during one frame
            ;when there are more voices used by the MIDI file than are
            ;actually available.

            shl   si,1
            mov   ax,[hwvtime+si]
            shr   si,1
            and   ax,ax
            jz    @time0
            cmp   ax,1
            jnz   @time1_n
            jmp   @time1
@time1_n:
            cmp   ax,2
            jz    @time2

            jmp   @nottime1
@time0:
            jmp   @skipvoice

@time2:
            ;set up remaining instrument data for frame number 2


            ;look up operator numbers

            mov   di,bx

            mov   dh,[si+vcop]                 ;DH = operator 2 index
            mov   dl,dh                        ;
            add   dl,3                         ;DL = operator 1 index

            mov   ah,es:[di+1+insdata+reg20]
            mov   al,20h                       ;base register
            add   al,dl                        ;operator 1
            adl_out                      ;

            add   al,20h
            mov   bl,dl
            xor   bh,bh
            mov   ah,[bx+hwreg40]
            adl_out

            ;set up waveform

            add  al,0a0h                        ;switch to wavef (reg. 0e0h)
            mov  ah,es:[di+insdata+wave_form_for_carrier]
            adl_out


            ;---------------------------------------
            ;set up hardware registers for modulator
            ;---------------------------------------

            ;this is a copy of the above code except for the operator selec-
            ;tion and the index of zero into instrument data instead of 1.

            mov   ah,es:[di+insdata+0+reg20]
            mov   al,20h                       ;base register
            add   al,dh                        ;operator 2
            adl_out                      ;

            add   al,20h
            mov   bl,dh
            mov   ah,[bx+hwreg40]
            adl_out

            ;set up waveform

            add  al,0a0h                       ;switch to wavefrm (reg. 0e0h)
            mov  ah,es:[di+insdata+wave_form_for_modulator]
            adl_out

            ;set up operator common parameters

            mov  al,0c0h                       ;feedback/connection
            add  ax,si
            mov  ah,es:[di+insdata+regc0]
            adl_out

            mov  bx,di
            jmp  @nottime1

@time1:

            ;set up instrument data for frame number 1

            ;look up operator numbers

            mov   di,bx

            mov   dh,[si+vcop]                 ;DH = operator 2 index
            mov   dl,dh                        ;
            add   dl,3                         ;DL = operator 1 index

;            mov   ah,es:[di+1+insdata+reg20]
;            mov   al,20h                       ;base register
;            add   al,dl                        ;operator 1
;            adl_out                      ;

;            add   al,20h
;            mov   bl,dl
            xor   bh,bh
;            mov   ah,[bx+hwreg40]
;            adl_out

            ;set up attack + initial decay rate

            mov   al,60h
            add   al,dl
            mov   ah,es:[di+insdata+1+reg60]     ;using same operator offset
            adl_out

            ;set up initial sustain level + release rate

            add   al,20h
            mov   ah,es:[di+insdata+1+reg80]   ;using same operator offset
            adl_out

            ;set up waveform

;            add  al,60h                        ;switch to wavef (reg. 0e0h)
;            mov  ah,es:[di+insdata+wave_form_for_carrier]
;            adl_out

            ;---------------------------------------
            ;set up hardware registers for modulator
            ;---------------------------------------

            ;this is a copy of the above code except for the operator selec-
            ;tion and the index of zero into instrument data instead of 1.

;            mov   ah,es:[di+insdata+0+reg20]
;            mov   al,20h                       ;base register
;            add   al,dh                        ;operator 2
;            adl_out                      ;

;            add   al,20h
;            mov   bl,dh
;            mov   ah,[bx+hwreg40]
;            adl_out

            ;set up attack + initial decay rate

            mov   al,60h
            add   al,dh
            mov   ah,es:[di+insdata+0+reg60]       ;using same operator offset
            adl_out

            ;set up initial sustain level + release rate

            add  al,20h
            mov   ah,es:[di+insdata+0+reg80]   ;using same operator offset
            adl_out

            ;set up waveform

;            add  al,60h                        ;switch to wavefrm (reg. 0e0h)
;            mov  ah,es:[di+insdata+wave_form_for_modulator]
;            adl_out

            ;set up operator common parameters

;            mov  al,0c0h                       ;feedback/connection
;            add  ax,si
;            mov  ah,es:[di+insdata+regc0]
;            adl_out

            mov  bx,di
            jmp  @skipvoice







@nottime1:
            ;
            ;Calculate modulation control's effect on voice LFO.
            ;This is a relatively complex task which involves
            ;various factors such as the key, the modulation
            ;sensitivity, the modulation rate and the logical position
            ;of the modulation wheel.
            ;

            mov   al,[hwvkey+si]                      ;key number
            xor   ah,ah
            shl   al,1                                ;as word index
            mov   di,ax                               ;key word index in DI
            mov   keyindex,ax                         ;save in local

            test  es:[bx+insdata+mod_lfo_rate],0ffh   ;Modulator LFO on?
            jnz   @mod_lfo
            jmp   @no_mod_lfo
@mod_lfo:
            test  [modvalue+si],0ffh
            jnz   @mod_on
            jmp   @no_mod_lfo
@mod_on:
            mov   ax,[fqdata+di+8]                    ;AX = fq of key+4
            sub   ax,[fqdata+di]                      ;AX = delta key fq
            cmp   ax,680                              ;roll octave?
            jl    @noroll1
            sub   ax,680
@noroll1:
            mov   cl,es:[bx+insdata+mod_lfo_depth]
            xor   ch,ch
            imul  cx                                  ;depth*value
            sar   ax,4
            mov   di,ax                               ;DI = modv

            mov   al,[hwvmodlfocnt+si]                ;modulation rate count
            cmp   al,es:[bx+insdata+mod_lfo_rate]             ;up or down?
            jb    @cnt_low
            mov   dl,es:[bx+insdata+mod_lfo_rate]             ;modd = rate*2 - modd
            shl   dl,1
            sub   dl,al
            mov   al,dl
@cnt_low:
            mov   dl,es:[bx+insdata+mod_lfo_rate]
            sar   dl,1
            sub   al,dl                               ;modd -= rate/2

            cbw
            imul  di                                  ;modd*modv

            mov   cl,es:[bx+insdata+mod_lfo_rate]
            xor   ch,ch
            idiv  cx                                  ;div modulation rate
nodivide:
            ;
            ;time to restart mod lfo count yet?
            ;

            mov   dl,es:[bx+insdata+mod_lfo_rate]
            shl   dl,1
            inc   [hwvmodlfocnt+si]
            cmp   [hwvmodlfocnt+si],dl
            jl    @notr1
            mov   [hwvmodlfocnt+si],0
@notr1:
            mov   cl,[modvalue+si]                    ;current mod position
            xor   ch,ch
            imul  cx                                  ;value * position
            sar   ax,7                                ;div by 128
            mov   modv,ax                             ;save as local
            jmp   short @mod_lfo_set
@no_mod_lfo:
            mov   modv,0                              ;clear mod lfo value
@mod_lfo_set:

            ;
            ;Calculate detune value of voice.
            ;NOTE: In fact this only needs to
            ;      be done once every note-on.
            ;      In effect this could reduce
            ;      the overall time spend by
            ;      the smfplay routine, but
            ;      not the maximum frame time.
            ;
            ;This is a fairly simple, and even
            ;rarely used function, so the amount
            ;so the of time saved by the suggested
            ;optimiziation is limited.
            ;

            test  es:[bx+insdata+detune],0ffh         ;any detune at all?
            jz    @no_detune
            mov   di,keyindex                         ;word index for hwvkey

            ;
            ;Value of detune equals delta(fq(note+1),fq(note))*detune
            ;                       ---------------------------------
            ;                                   detune_max

            mov   ax,[di+fqdata+2]
            sub   ax,[di+fqdata]
            cmp   ax,680
            jl    @lt1
            sub   ax,680
@lt1:
            xor   dh,dh
            mov   dl,es:[bx+insdata+detune]
            mul   dx

            shr   ax,5                                ;detune_max = 64

            mov   detunev,ax                          ;save as local
            jmp   short @detune_set
@no_detune:
            mov   detunev,0                           ;save as local
@detune_set:

            shl   si,1                                ;SI changed to word idx

            ;
            ;And as the last frequency-affecting parameter, calculate
            ;the time variable pitch (tvp) parameter.
            ;

            cmp   es:[bx+insdata+tvp_depth],32                ;depth == zero?
            jz    @tvp_none

            mov   ax,[si+hwvtime]
            and   ah,ah                               ;end of tvp affection?
            jnz   @tvp_none
            cmp   al,es:[bx+insdata+tvp_rate]
            jae   @tvp_none                           ;yes, don't calculate

            mov   ax,0400h-680
            mov   dl,es:[bx+insdata+tvp_depth]
            xor   dh,dh
            sub   dx,32
            imul  dx
            sar   ax,5                                ;AX = rate at note-on

            mov   dl,es:[bx+insdata+tvp_rate]                 ;rate
            sub   dl,byte ptr [si+hwvtime]            ;- time
            xor   dh,dh
            imul  dx                                  ;rate * current

            mov   cl,es:[bx+insdata+tvp_rate]
            xor   ch,ch
            idiv  cx                                  ;/progress
@nodivide:
            jmp   @tvp_set
@tvp_none:
            xor   ax,ax
@tvp_set:

            ;
            ;Now all that's left is to let the changes reflect on the
            ;frequency value of the voice and update the hardware regs.
            ;

            mov   di,[si+hwvfq]                       ;SI still word index
            add   di,ax                               ;TVP value from AX

            ;examine if frequency change requires an octave shift

            mov   dx,di
            and   dx,03ffh
            cmp   dx,688
            jl    @l6881
            add   di,680
            jmp   short @a3441
@l6881:
            cmp   dx,344
            jg    @a3441
            sub   di,680
@a3441:

            add   di,modv                             ;modulation

            mov   dx,di
            and   dx,03ffh
            cmp   dx,688
            jl    @l6882
            add   di,680
            jmp   short @a3442
@l6882:
            cmp   dx,344
            jg    @a3442
            sub   di,680
@a3442:

            add   di,detunev                          ;detune

            mov   dx,di
            and   dx,03ffh
            cmp   dx,688
            jl    @l6883
            add   di,680
            jmp   short @a3443
@l6883:
            cmp   dx,344
            jg    @a3443
            sub   di,680
@a3443:

            shr   si,1                                ;SI back to byte index
            mov   dx,di

            mov   ah,dl
            mov   al,0a0h
            add   ax,si                               ;voice+reg 0a0h
            adl_out                                   ;output lo value

            add   al,10h                              ;register 0b0h+voice
            mov   ah,dh                               ;hi value

            or    ah,20h
            adl_out
@skipvoice:
            shl   si,1
            inc   [hwvtime+si]
            shr   si,1
@@skipvoice:
            inc   si                                  ;increment byte index
            cmp   si,9                                ;9 voices
            jae   @voice_loop_end
            jmp   @voice_loop

@voice_loop_end:
            ret
hw_isr      endp

end         hw_entry_point
;end of file
